/*
** Copyright  1993, 1994 by Eric M. LaFranchi.  All Rights Reserved.
**
** This software is Copyright 1993, 1994 by Eric M. LaFranchi.  Permission
** to use, copy, and freely redistributed this software in its entirety is
** hereby granted provided that the above copyright notice and this
** permission notice are retained.  This software may not be sold for profit
** or incorporated in commercial software products without the written
** permission of the author.  This software is provided "as is", the author
** nor his employer make any representation of warranty, express or implied,
** with respect to any code or other information herein.  In addition, the
** author disclaim's any liability whatsoever for any use of such code or
** other information.
** 
**+-+
**
**
** Module: VMS_MISC
**
** Abstract:
**	This module includes all the functions that access VMSmail.
**	These functions provide a clean interface to perform mail file
**	access.
**
** Author:
**	Eric M. LaFranchi
**
** Creation Date:
**	18-APR-1993
**
** Modifications History:
**	EML001		Eric M. LaFranchi	14-Jan-1994
**	Added support for selecting the mailfile to read.
**
**-+-
*/

#include <ctype.h>			/* */
#include <stdio.h>			/* Standard I/O definitions.	     */
#include <stdlib.h>			/* Standard library definitions.     */
#include <string.h>			/* Standard string definitions.      */
#include <maildef.h>			/* mail service definitions	     */
#include <descrip.h>			/* VMS descriptor definitions	     */
#include <ssdef.h>			/* Standard VMS messages code	     */
#include "vmsdef.h"
#include "msgdef.h"
#include "packdef.h"

    /* define mail messages codes not in maildef.h. Values taken from
     * sys$library starlet.req
     */
#ifndef MAIL$_MSGTEXT
#define MAIL$_MSGTEXT 8322305
#endif

#ifndef MAIL$_NOMOREREC
#define MAIL$_NOMOREREC 8314792
#endif

#ifndef MAIL$_NOMOREMSG
#define MAIL$_NOMOREMSG	8290386
#endif

#ifndef MAIL$_NOTEXIST
#define MAIL$_NOTEXIST	8290522
#endif

#define MAIL_C_BUFSIZ 256

    /* extern function declaration
     */
void	LIB$ESTABLISH( const size_t (*)( ) );

size_t	handler( struct chf$signal_array *signal_array,
		 struct chf$mech_array *mech_array );

size_t	MAIL$MAILFILE_BEGIN( void **, const ITMLST * const , const ITMLST * const );
size_t	MAIL$MAILFILE_OPEN( void **, const ITMLST * const, const ITMLST * const );
size_t	MAIL$MAILFILE_CLOSE( void **, const ITMLST * const, const ITMLST * const );
size_t	MAIL$MAILFILE_END( void **, const ITMLST * const, const ITMLST * const );
size_t	MAIL$MESSAGE_BEGIN( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$MESSAGE_DELETE( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$MESSAGE_COPY( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$MESSAGE_INFO( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$MESSAGE_SELECT( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$MESSAGE_GET( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$MESSAGE_END( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$USER_BEGIN( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$USER_SET_INFO( void **, const ITMLST * const, const ITMLST * const ); 
size_t	MAIL$USER_END( void **, const ITMLST * const, const ITMLST * const ); 

struct context
{
    void *mfcontext;			/* mail file context pointer	*/
    void *msgcontext;			/* message context pointer	*/
    unsigned int article_count;		/* article count in folder      */
};

#ifdef VAXC
#pragma nostandard
#endif
extern size_t (*pack_connect)( void **const, const char *const );
extern size_t (*select_articles)( void *const, const char *const );
extern size_t (*scan_articles)( void *const, size_t (*)(), const void * );
extern size_t (*extract_articles)( void *const, const size_t, size_t (*)(), const void * );
extern size_t (*cleanup_articles)( void *const, const size_t );
extern size_t (*disconnect)( void **const );

extern const unsigned int gblflags;

#ifdef VAXC
#pragma standard
#endif

static const ITMLST nullitem[ ] = { 0, 0, NULL, NULL };


/*
 *+-+
 *
 * Function: mail_open
 *
 * Abstract:
 *	mail_open localizes the opening of the mail file and sets up the
 *	context for message processing.
 *
 * Inputs:
 *	context -- address of a pointer to structure that will contains mail
 *		   file and message context information.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_open( void **const context,
	   const char *const mailfile )
{
    static ITMLST in_itmlst[ ] = {
	{ sizeof( struct context * ), MAIL$_MESSAGE_FILE_CTX, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    static ITMLST open_itmlst[ ] = {
	{ 0, MAIL$_MAILFILE_NAME, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)malloc( sizeof( *cxt ) );
    if ( cxt == NULL )
	raise_exception( PACKASM_INSVIRMEM, 0 );

    status = MAIL$MAILFILE_BEGIN( &cxt->mfcontext, nullitem, nullitem );
    if ( !(status & 1) ) return ( status );

    if ( mailfile && mailfile[0] != '\0' )
    {
	open_itmlst[0].buflen = strlen( mailfile );
	open_itmlst[0].bufadr = (char *)mailfile;
    }

    status = MAIL$MAILFILE_OPEN( &cxt->mfcontext, open_itmlst, nullitem );
    if ( !(status & 1) ) return ( status );

    in_itmlst[0].bufadr = &cxt->mfcontext;
    status = MAIL$MESSAGE_BEGIN( &cxt->msgcontext, in_itmlst, nullitem );
    if ( !(status & 1) ) return ( status );

    cxt->article_count = 0;
    *context = cxt;

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_select_folder
 *
 * Abstract:
 *	This function selects the specified mail folder.
 *
 * Inputs:
 *	context -- address of a pointer to structure that contains mail file
 *		   and message context information
 *
 *	folder -- pointer to folder name.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_select_folder( void *const context,
		    const char *const folder )
{
    static ITMLST folderitem[ ] = {
	{ 0, MAIL$_MESSAGE_FOLDER, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    static ITMLST msgcntitem[ ] = {
	{ sizeof( unsigned int ), MAIL$_MESSAGE_SELECTED, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    unsigned int msgcnt;
    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

	/* fill in folders and message count descriptors
	 */
    msgcntitem[0].bufadr = &msgcnt;
    folderitem[0].buflen = strlen( folder );
    folderitem[0].bufadr = (char *)folder;

    status = MAIL$MESSAGE_SELECT( &cxt->msgcontext, folderitem, msgcntitem );
    if ( !(status & 1) )
	return ( status );

    cxt->article_count = msgcnt;

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_scan_articles
 *
 * Abstract:
 *	This routine scans the specified VMSmail folder for subject lines
 *	that have packages that match the specified package name and other
 *	specified criteria. If a match is found the folder and article number
 *	passed to the add_part_record function.
 *
 * Inputs:
 *	context -- pointer to structure that will contains mail file
 *		   and message context information.
 *	packname -- pointer to package to compare against
 *	before -- pointer to user specified VMS time
 *	since -- pointer to user specified VMS time
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_scan_articles( void *const context,
		    size_t (*const callback)( ),
		    const void *callarg )
{
    static ITMLST msgiditem[ ] = {
	{ sizeof( unsigned int ), MAIL$_MESSAGE_ID, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    static ITMLST infoitem[ ] = {
	{ 0, MAIL$_MESSAGE_SUBJECT, NULL, NULL },
	{ 0, MAIL$_MESSAGE_BINARY_DATE, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    char subject[MAIL_C_BUFSIZ];
    unsigned int msgid, sublen, date[2];

    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

	/* item to be selected
	 */
    msgiditem[0].bufadr = &msgid;

    infoitem[0].buflen = sizeof( subject ) - 1;
    infoitem[0].bufadr = subject;
    infoitem[0].retlen = &sublen;
    infoitem[1].bufadr = date;
    infoitem[1].buflen = sizeof( date );

	/* select each message in the folder, then * try to determine
	 * if it this part of the package we are trying to extract.
	 */
    for ( msgid = 1; msgid <= cxt->article_count; msgid++ )
    {
	status = MAIL$MESSAGE_INFO( &cxt->msgcontext, msgiditem, infoitem );
	if ( !(status & 1) )
	    return ( status );

	subject[sublen] = '\0';

	status = callback( callarg, subject, msgid, date );
	if ( !(status & 1) )
	    continue;
    }

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_extract_article
 *
 * Abstract:
 *	This function locates specified part using the folder and msgid
 *	information specified by the user.
 *
 * Inputs:
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_extract_article( void *const context,
		      const unsigned int msgid,
		      size_t (*const callback)( ),
		      const void *callarg )
{
    static ITMLST msgiditem[ ] = {
	{ sizeof( unsigned int ), MAIL$_MESSAGE_ID, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    static ITMLST continueitem[ ] = {
	{ 0, MAIL$_MESSAGE_CONTINUE, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    static ITMLST recorditem[ ] = {
	{ 0, MAIL$_MESSAGE_RECORD, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    char mail_record[MAIL_C_BUFSIZ];
    unsigned int record_length;
    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

	/* select message
	 */
    msgiditem[0].bufadr = (unsigned int *)&msgid;
    status = MAIL$MESSAGE_GET( &cxt->msgcontext, msgiditem, nullitem );
    if ( !(status & 1) )
	return ( status );

    recorditem[0].buflen = sizeof( mail_record );
    recorditem[0].bufadr = mail_record;
    recorditem[0].retlen = &record_length;

	    /* read records from mail file and write to VMS file.
	     */
    for ( ;; )
    {
	status = MAIL$MESSAGE_GET( &cxt->msgcontext, continueitem, recorditem );

	if ( status == MAIL$_NOMOREREC )
	    break;
	
	if ( !(status & 1) )
	    return ( status );

	mail_record[record_length] = '\0';

	status = callback( callarg, mail_record );
	if ( !(status & 1) )
	    return (status );
    }

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_delete_article
 *
 * Abstract:
 *	this routine delete the specified mail article from the
 *	previously selected folder.
 *
 * Inputs:
 *	context -- address of a pointer to structure that contains mail
 *		   file and message context information.
 *	msgid -- article identification for the article to be deleted
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_delete_article( void *const context,
		     const unsigned int msgid )
{
    static ITMLST msgiditem[ ] = {
	{ sizeof( size_t ), MAIL$_MESSAGE_ID, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

	/* article to be deleted
	 */
    msgiditem[0].bufadr = (unsigned int *)&msgid;

    status = MAIL$MESSAGE_DELETE( &cxt->msgcontext, msgiditem, nullitem );
    if ( !(status & 1) )
	return (status );

    return ( SS$_NORMAL );
}


/*
 *+-+
 *
 * Function: mail_copy_article
 *
 * Abstract:
 *	this routine copies a message from the currently select folder
 *	to the specified folder
 *
 * Inputs:
 *	context -- address of a pointer to structure that will contains mail
 *		   file and message context information.
 *	msgid -- article identification for the article to be deleted
 *	folder -- pointer to folder name of the folder to which the
 *		  message is to be copied.
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_copy_article( void *const context,
		   const unsigned int msgid,
		   const char *const folder )
{
    static ITMLST copyitem[ ] = {
	{ sizeof( size_t ), MAIL$_MESSAGE_ID, NULL, NULL },
	{ 0, MAIL$_MESSAGE_FOLDER, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

	/* article to be copied
	 */
    copyitem[0].bufadr = (unsigned int *)&msgid;
    copyitem[1].buflen = strlen( folder );
    copyitem[1].bufadr = (char *)folder;

    status = MAIL$MESSAGE_COPY( &cxt->msgcontext, copyitem, nullitem );
    if ( !(status & 1) )
	return (status );

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_move_article
 *
 * Abstract:
 *	this routine copies a message from the currently select folder
 *	to the specified folder
 *
 * Inputs:
 *	context -- address of a pointer to structure that will contains mail
 *		   file and message context information.
 *	msgid -- article identification for the article to be deleted
 *	folder -- pointer to folder name of the folder to which the
 *		  message is to be copied.
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_move_article( void *const context,
		   const unsigned int msgid,
		   const char *const folder )
{
    register size_t status;

	/* copy article to specified folder
	 */
    status = mail_copy_article( context, msgid, folder );
    if ( !(status & 1) )
	return (status );

	/* delete the specified article from the selected folder
	 */
    status = mail_delete_article( context, msgid );
    if ( !(status & 1) )
	return (status );

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_newmail_count
 *
 * Abstract:
 *	This function sets the correct message count in the new mail
 *	folder.
 *
 * Inputs:
 *	context -- address of a pointer to structure that contains mail file
 *		   and message context information
 *
 *	folder -- pointer to folder name.
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_newmail_count( void *const context )
{
    static ITMLST folderitem[ ] = {
	{ 0, MAIL$_MESSAGE_FOLDER, NULL, NULL },
	{ 0, MAIL$_NOSIGNAL, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    static ITMLST msgcntitem[ ] = {
	{ sizeof( size_t ), MAIL$_MESSAGE_SELECTED, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    static ITMLST newcntitem[ ] = {
	{ 0, MAIL$_USER_USERNAME, NULL, NULL },
	{ sizeof( unsigned short ), MAIL$_USER_SET_NEW_MESSAGES, NULL, NULL },
	{ 0, MAIL$_NOSIGNAL, NULL, NULL },
	{ 0, 0, NULL, NULL } };

    void *lcontext = NULL;
    char username[12];
    unsigned int msgcnt;

    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)context;

	/* Use default NEWMAIL folder
	 */
    msgcntitem[0].bufadr = &msgcnt;
    folderitem[0].buflen = sizeof( "NEWMAIL" ) - 1;
    folderitem[0].bufadr = "NEWMAIL";

	/* select the folder "NEWMAIL". Assume that an error means that
	 * the folder doesn't exist, so there's a zero message count.
	 */
    status = MAIL$MESSAGE_SELECT( &cxt->msgcontext, folderitem, msgcntitem );
    if ( !(status & 1) )
    {
	if ( status != MAIL$_NOTEXIST )
	    return ( status );
	msgcnt = 0;
    }

    cxt->article_count = msgcnt;

    status = get_username( username, sizeof( username ) );
    if ( !(status & 1) )
	raise_exception( PACKASM_INTERNERR, 0, status );

    status = MAIL$USER_BEGIN( &lcontext, nullitem, nullitem );
    if ( !(status & 1) )
	 return ( status );

    newcntitem[0].buflen = sizeof( username );
    newcntitem[0].bufadr = username;
    newcntitem[1].bufadr = &msgcnt;

    status = MAIL$USER_SET_INFO( &lcontext, newcntitem, nullitem );
    if ( !(status & 1) )
	return ( status );

    status = MAIL$USER_END( &lcontext, nullitem, nullitem );
    if ( !(status & 1) )
	return ( status );

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_close
 *
 * Abstract:
 *	close_mail localizes the releasing of mail file message context
 *	and close the mail file.
 *
 * Inputs:
 *	context -- address of a pointer to structure that contains mail file
 *		   and message context information
 *
 * Outputs:
 *	None
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
static size_t
mail_close( void **const context )
{
    register struct context *cxt;
    register size_t status;

    LIB$ESTABLISH( handler );

    cxt = (struct context *)*context;

    status = mail_newmail_count( cxt );
    if ( !(status & 1) ) return ( status );

    status = MAIL$MESSAGE_END( &cxt->msgcontext, nullitem, nullitem  );
    if ( !(status & 1) ) return ( status );

    status = MAIL$MAILFILE_CLOSE( &cxt->mfcontext, nullitem, nullitem );
    if ( !(status & 1) ) return ( status );

    status = MAIL$MAILFILE_END( &cxt->mfcontext, nullitem, nullitem );
    if ( !(status & 1) ) return ( status );

    *context = NULL;
    free( cxt );

    return ( SS$_NORMAL );
}

/*
 *+-+
 *
 * Function: mail_load_addresses
 *
 * Abstract:
 *	this routine loads the addresses of all the mail routines in
 *	the global data structures.
 *
 * Inputs:
 *	None
 *
 * Outputs:
 *	global function	pointer to mail routines loaded
 *
 * Returns:
 *	returns VMS status error or SS$_NORMAL
 *
 *-+-
 */
size_t
load_VMSmail_routines( )
{
    pack_connect = mail_open;
    select_articles = mail_select_folder;
    scan_articles = mail_scan_articles;
    extract_articles = mail_extract_article;
    cleanup_articles = mail_delete_article;
    disconnect = mail_close;

    return ( PACKASM_SUCCESS );
}
